#include <exception>
#include <string>

struct Utf8Error : std::exception {
    char const *what() throw() {
        return "UTF8 error";
    }
};

class PctUtf8ToLatin1 {
    
public:

    PctUtf8ToLatin1(char const *input)
    :   ptr((unsigned char const *) input) {
    }

    bool convert() {
        try {
            while (convert_seq()) {
                // empty
            }
        } catch (Utf8Error const &) {
            return false;
        }
        return true;
    }

    char const *result() const {
        return out.c_str();
    }

private:

    void checkIsFollowByte(int const v) {
        if ((v & 0xC0) != 0x80)
            throw Utf8Error();
    }

    bool convert_seq() {
        int const c = *ptr;
        if (c == 0)
            return false;
        if (c != '%') {
            out += (char) c;
            ++ ptr;
            return true;
        }
        int const n = read_utf8_seq();
        if (n <= 255) {
            char buf[8];
            sprintf(buf, "%%%02X", n);
            out += buf;
        } else {
            out += "%3F"; // ?
        }
        return true;
    }

    int read_utf8_seq() {
        int const pctValue = readPctValue();
        if (pctValue < 0x80)
            return pctValue;
        if (pctValue < 0xC0)
            throw Utf8Error();
        if (pctValue < 0xE0) {
            // 2-Byte UTF-8 Sequenz
            int const pctValue2 = readPctValue();
            checkIsFollowByte(pctValue2);
            int const v = ((pctValue & 0x1F) << 6) | (pctValue2 & 0x3F);
            if (v < 0x80) {
                // Illegal: UTF-8 verlangt, dass die kuerzeste moegliche
                // Sequenz verwendet wird.
                throw Utf8Error();
            }
            return v;
        }
        if (pctValue < 0xF0) {
            // 3-Byte UTF-8 Sequenz
            for (int n = 0; n < 2; ++n)
                checkIsFollowByte(readPctValue());
            return '?';
        }
        if (pctValue < 0xF8) {
            // 4-Byte UTF-8 Sequenz
            for (int n = 0; n < 3; ++n)
                checkIsFollowByte(readPctValue());
            return '?';
        }
        // 5 oder mehr Bytes UTF-8: Kein Unicode-Zeichen.
        throw Utf8Error();
    }

    int readHexDigit() {
        int const c = *ptr;
        int res;
        if (c >= '0' && c <= '9')
            res = c - '0';
        else if (c >= 'A' && c <= 'F')
            res = c - 'A' + 10;
        else if (c >= 'a' && c <= 'f')
            res = c - 'a' + 10;
        else
            throw Utf8Error();
        ++ptr;
        return res;
    }

    int readPctValue() {
        if (*ptr != '%')
            throw Utf8Error();
        ++ ptr;
        int h1, h2;
        h1 = readHexDigit();
        h2 = readHexDigit();
        return h1 * 16 + h2;
    }

    std::string out;
    unsigned char const *ptr;
};
