/*
BOIL (Brunni's Own Interpreter Language)
(c) 1998 netEstate (http://www.netestate.de/, info@netestate.de),
Michael Brunnbauer & Markus Hendel, Gesellschaft fuer Netzwerkkommunikation bR
Am Bluetenanger 8, 80995 Muenchen, Germany

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "general.h"
#ifdef WITH_RPC
#include "rpc.h"
#endif
#include "handles.cc"
#include "boil.h"
#include "boil.tab.h"
#include "libs.h"

#include "PostdataBuilder.h"

struct mimetype
{
  char* ext;
  char* type;
};
 
mimetype mimetypes[]=
{
  { "gif","image/gif" },
  { "jpg","image/jpeg" },
  { "jpeg","image/jpeg" },
  { "jpe","image/jpeg" },
  { "html","text/html" },
  { "htm","text/html" },
  { "pdf","application/pdf" },
  { "zip","application/zip" },
  { "xls","application/vnd.ms-excel" },
  { "","application/octet-stream" },
  { 0,"application/octet-stream" }
};

char* cgiext[]=
{
  "cgi",0
};

int args_cgiwrap[]= { TYPE_VALUE_STRING,
                      TYPE_VALUE_STRING,
                      TYPE_VALUE_ULONG,
                      0 };

#include "PctUtf8ToLatin1.h"

flowctrl func_cgiwrap(type_block* aktblock,opt* options,
                      joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=0;
  char buf[CGIWRAP_BUF];

  struct stat filestat;

  if (!defined((*args)[0].stringval) ||
      !defined((*args)[1].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","cgiwrap: string not defined");
      return back;
    };

  unsigned long nohtmlheader=(*args)[2].val.ulongval;
  string file=(*args)[0].stringval;

  if (strstr(file.c_str(),".."))
    {
      setvarulong("error",1);
      setvarstring("perror","cgiwrap: security violation '..' in "+file);
      return back;
    };

  if (stat(file.c_str(),&filestat)==-1)
    {
      setvarulong("error",1);
      setvarstring("perror","cgiwrap: cannot stat "+file);
      return back;
    };
  if (S_ISDIR(filestat.st_mode))
    {
      if (!stat((file+"/index.html").c_str(),&filestat))
        {
          file+="/index.html";
        }
      else if (!stat((file+"/index.htm").c_str(),&filestat))
        {
          file+="/index.htm";
        }
      else
        {
          setvarulong("error",1);
          setvarstring("perror","cgiwrap: index.htm[l] not found: "+file);
          return back;
        };
    }
  else if (!S_ISREG(filestat.st_mode) && !S_ISLNK(filestat.st_mode))
    {
      setvarulong("error",1);
      setvarstring("perror","cgiwrap: special file: "+file);
      return back;
    };

  unsigned long i;
  char* empty=".";
  char* extension=strrchr(file.lock(),(int) '.');
  if (!extension)
    {
      extension=empty;
    };
  extension++;
  string extension1=extension;
  str_lwr(extension1.lock());

  /* cgi-skript ? */
  for (i=0;cgiext[i];i++)
    {
      if (!strcmp(cgiext[i],extension1.c_str()))
        {
          char* dummy=0;char** argv=&dummy;
          if ((*args)[1].stringval!="")
            {
              if (setenv("REQUEST_METHOD","POST",1)==-1)
                {
                  setvarulong("error",1);
                  setvarstring("perror",
                               "cgiwrap: cannot setenv(REQUEST_METHOD)");
                  return back;
                };
            };
          FILE* prg=myexec(file.lock(),argv);
          if (!prg)
            {
              setvarulong("error",1);
              setvarstring("perror",
                           "cgiwrap: cannot exec "+(*args)[0].stringval);
              return back;
            };
          if ((*args)[1].stringval!="")
            {
              if (fwrite((*args)[1].stringval.c_str(),1,(*args)[1].stringval.size(),
                         prg)!=(*args)[1].stringval.size())
                {
                  setvarulong("error",1);
                  setvarstring("perror","cgiwrap: could not write postdata");
                  return back;
                };
            };
          if (fclose(prg))
            {
              setvarulong("error",1);
              setvarstring("perror","cgiwrap: returncode of fclose()!=0");
              return back;
            };
          back.returnval.val.ulongval=1;
          return back;
        };
    };

  /* search content type */
  string ctype=0;
  for (i=0;mimetypes[i].ext;i++)
    {
      if (!strcmp(mimetypes[i].ext,extension1.c_str()))
        {
          ctype=mimetypes[i].type;
          break;
        };
    };
  if (!defined(ctype)) { ctype=mimetypes[i].type; };
  
  int handle=open(file.c_str(),O_RDONLY);
  if (handle==-1)
    {
      setvarulong("error",1);
      setvarstring("perror","cgiwrap: cannot open "+file);
      return back;
    };
  
  /* get file size */
  struct stat stats;
  if (stat(file.c_str(),&stats)==-1)
    {
      setvarulong("error",1);
      setvarstring("perror","cgiwrap: cannot stat "+file);
      return back;
    };
  size_t filesize=stats.st_size;

  joined_list<string> ranges;
  joined_list<unsigned long> ranges_start;
  joined_list<unsigned long> ranges_len;

  string range=getenv("HTTP_RANGE");
  if (defined(range))
    {
      unsigned long dummy;

      strtok(range.lock(),"=");
      string token;
      while (defined(token=strtok(0,","))) ranges.insert_listelem(token);

      for (i=0;i!=ranges.length();i++)
	{
	  string r=ranges[i];
	  if (r[0]=='-')
	    {
	      unsigned long len=atol(r.c_str()+1);
	      if (len>filesize) break;

	      dummy=filesize-len;
	      ranges_start.insert_listelem(dummy);
	      ranges_len.insert_listelem(len);
	    }
	  else
	    {
	      unsigned long start=atol(strtok(r.lock(),"-"));
	      string e=strtok(0,"-");
	      if (defined(e) && e!="")
		{
		  if (start>=filesize) break;
		  ranges_start.insert_listelem(start);
		  
		  unsigned long ende=atol(e.c_str());
		  if (ende<start || ende>=filesize)
		    {
		      ranges_start.remove_listelem(0);
		      break;
		    };
		  
		  dummy=ende-start+1;
		  ranges_len.insert_listelem(dummy);
		}
	      else
		{
		  if (start>=filesize) break;
		  ranges_start.insert_listelem(start);
		  dummy=filesize-start;
		  ranges_len.insert_listelem(dummy);
		};
	    };
	};
    };

  if (nohtmlheader) 
    {
      while (ranges_start.length()) ranges_start.remove_listelem(0);
      while (ranges_len.length()) ranges_len.remove_listelem(0);
    };

  if (ranges_start.length())
    {
      printf("%s","Status: 206 Partial Content\r\nAccept-ranges: bytes\r\nContent-type: multipart/x-byteranges; boundary=THIS_STRING_SEPARATES\r\n\r\n");
      fflush(stdout);

      while (ranges_start.length())
	{
	  unsigned long start=ranges_start[0];
	  unsigned long len=ranges_len[0];
	  ranges_start.remove_listelem(0);
	  ranges_len.remove_listelem(0);

	  printf("\r\n--THIS_STRING_SEPARATES\r\nContent-type: %s\r\nContent-Range: bytes %lu-%lu/%lu\r\n\r\n", ctype.c_str(),start,start+len-1,filesize);
	  fflush(stdout);

	  lseek(handle,start,SEEK_SET);
	  while (len)
	    {
	      size_t size;
	      size=read(handle,buf,CGIWRAP_BUF);
	      if (size>len) size=len;
	      if ((unsigned long) write(1,buf,size)!=
		  (unsigned long) size)
		{
		  setvarulong("error",1);
		  setvarstring("perror",
			       "cgiwrap: output of "+file+" failed");
		  close(handle);
		  return back;
		};
	      len-=size;
	    };
	};
      printf("\r\n--THIS_STRING_SEPARATES--\r\n");
      fflush(stdout);
    }
  else
    {
      if (!nohtmlheader)
	{
	  printf("Accept-ranges: bytes\r\nContent-type: %s\r\nContent-Length: %lu\r\n\r\n", ctype.c_str(),(unsigned long) filesize);
	  fflush(stdout);
	};

      size_t size;
      while ((size=read(handle,buf,CGIWRAP_BUF))>0)
	{
	  if ((unsigned long) write(1,buf,size)!=
	      (unsigned long) size)
	    {
	      setvarulong("error",1);
	      setvarstring("perror","cgiwrap: output of "+file+" failed");
	      close(handle);
	      return back;
	    };
	};
      close(handle);
    };

  back.returnval.val.ulongval=1;
  return back;
}

int args_getpost[]= { 0 };

flowctrl func_getpost(type_block* aktblock,opt* options,
                      joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back,call,call1;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=0;

  char buf[MAX_LINE_SIZE];
  string postdata="";
  char* ptr;
  size_t size;
  int status;
  type_value var;
  type_value* varptr;
  var.type=TYPE_VALUE_STRING;
  hashdir<char,string,256,0> declared;

  if (!getenv("REQUEST_METHOD"))
    {
      setvarulong("error",1);
      setvarstring("perror",
                   "getpost: environment-variable REQUEST_METHOD not defined");
      return back;
    };

  if (strcmp(getenv("REQUEST_METHOD"),"POST"))
    {
      setvarulong("error",1);
      setvarstring("perror","getpost: request-method is not POST");
      return back;
    };

  while ((size=fread(buf,1,MAX_LINE_SIZE-1,stdin))==MAX_LINE_SIZE-1)
    {
      buf[size]=0;
      postdata+=buf;
    };
  if (size) { buf[size]=0; postdata+=buf; };

  ptr=postdata.lock();
  
  back.returnval.val.ulongval=1;
  string name;
  while ((status=urlencode_parse(&ptr,name,var.stringval))==1)
    {
      call=declarevar(aktblock,options,&var,name);
      if (call.ctrl!=FLOW_OK)
        {
          /* already declared by getpost() ? */
          if (declared.get_listelem(name))
            {
              /* yes -> append value */
              call1=getvar(aktblock,options,varptr,name);
              if (call1.ctrl!=FLOW_OK)
                {
                  setvarulong("error",1);
                  setvarstring("perror",
                               "getpost: hum ? could not get already\
defined variable "+name+":"+call1.error);
                  back.returnval.val.ulongval=0;
                }
              else
                {
                  varptr->stringval+=","+var.stringval;
                };
            }
          else
            {
              setvarulong("error",1);
              setvarstring("perror","getpost: could not declare variable "+
                           name+":"+call.error);
              back.returnval.val.ulongval=0;
            };
        }
      else
        {
          char ch;
          declared.insert_listelem(ch,name);
        };
    };
  
  if (status==-1)
    {
      setvarulong("error",1);
      setvarstring("perror","getpost: syntax error in post-data");
      back.returnval.val.ulongval=0;
    }
  else
    {
      call=declarevar(aktblock,options,&var,name);
      if (call.ctrl!=FLOW_OK)
        {
          if (declared.get_listelem(name))
            {
              /* yes -> append value */
              call1=getvar(aktblock,options,varptr,name);
              if (call1.ctrl!=FLOW_OK)
                {
                  setvarulong("error",1);
                  setvarstring("perror",
                               "getpost: hum ? could not get already defined variable "+name+":"+call1.error);
                  back.returnval.val.ulongval=0;
                }
              else
                {
                  varptr->stringval+=","+var.stringval;
                };
            }
          else
            {
	      setvarulong("error",1);
	      setvarstring("perror",
			   "getpost: could not declare variable "+name+":"+
			   call.error);
	      back.returnval.val.ulongval=0;
	    };
        };
    };
  
  return back;
}

int args_postdata_utf8tolatin1[]= { TYPE_VALUE_STRING,
                                    0 };

flowctrl func_postdata_utf8tolatin1(type_block *aktblock,
				    opt *options,
				    joined_list<type_value> *args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back,call,call1;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","postdata_utf8tolatin1: string not defined");
      return back;
    };

  PctUtf8ToLatin1 converter((*args)[0].stringval.c_str());
  if (!converter.convert())
    {
      back.returnval.stringval=(*args)[0].stringval;
      setvarulong("error",1);
      setvarstring("perror","postdata_utf8tolatin1: illegal utf8");
      return back;
    };
  back.returnval.stringval=converter.result();
  return back;
}


int args_postdata[]= { TYPE_VALUE_STRING,
                       0 };

flowctrl func_postdata_inner(type_block *aktblock,
                             opt *options,
                             joined_list<type_value> *args,
                             bool utf8)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back,call,call1;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=0;
  char *ptr;
  int status;
  type_value var;
  type_value* varptr;
  var.type=TYPE_VALUE_STRING;
  hashdir<char,string,256,0> declared;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","postdata: string not defined");
      return back;
    };

  char *raw_ptr=(*args)[0].stringval.lock();
  
  string latin1;
  if (utf8) {
    PctUtf8ToLatin1 converter(raw_ptr);
    if (!converter.convert()) {
      setvarulong("error",1);
      setvarstring("perror", "postdata: illegal utf8");
      back.returnval.val.ulongval=0;
      return back;
    }
    latin1=converter.result();
    ptr=latin1.lock();
  } else {
    ptr = raw_ptr;
  }

  back.returnval.val.ulongval=1;
  string name;

  while (1)
    {
      status=urlencode_parse(&ptr,name,var.stringval);
      if (status!=1) break;

      call=declarevar(aktblock,options,&var,name);
      if (call.ctrl!=FLOW_OK)
        {
          if (declared.get_listelem(name))
            {
              /* yes -> append value */
              call1=getvar(aktblock,options,varptr,name);
	      if (call1.ctrl!=FLOW_OK)
		{
		  setvarulong("error",1);
		  setvarstring("perror",
			       "postdata: hum ? could not get already\
defined variable "+name+":"+call1.error);
		  back.returnval.val.ulongval=0;
		}
	      else
		{
		  varptr->stringval+=","+var.stringval;
		};
            }
          else
            {
              setvarulong("error",1);
              setvarstring("perror","postdata: could not declare variable "+
                           name+":"+call.error);
              back.returnval.val.ulongval=0;
            };
        }
      else
        {
          char ch;
          declared.insert_listelem(ch,name);
        };
    };

  if (status==-1)
    {
      setvarulong("error",1);
      setvarstring("perror","postdata: syntax error in post-data");
      back.returnval.val.ulongval=0;
    }
  else
    {
      call=declarevar(aktblock,options,&var,name);
      if (call.ctrl!=FLOW_OK)
        {
          if (declared.get_listelem(name))
            {
              /* yes -> append value */
              call1=getvar(aktblock,options,varptr,name);
	      if (call1.ctrl!=FLOW_OK)
		{
		  setvarulong("error",1);
		  setvarstring("perror",
			       "postdata: hum ? could not get already defined variable "+name+":"+call1.error);
		  back.returnval.val.ulongval=0;
		}
	      else
		{
		  varptr->stringval+=","+var.stringval;
		};
            }
          else
            {
	      
	      setvarulong("error",1);
	      setvarstring("perror",
			   "postdata:could not declare variable "+name+":"+
			   call.error);
	      back.returnval.val.ulongval=0;
	    };
        };
    };
  
  return back;
}

flowctrl func_postdata(type_block *aktblock,
                       opt *options,
                       joined_list<type_value> *args) {
    return func_postdata_inner(aktblock, options, args, false);
}

flowctrl func_postdata_utf8(type_block *aktblock,
                            opt *options,
                            joined_list<type_value> *args) {
    return func_postdata_inner(aktblock, options, args, true);
}

int args_file_upload[] = { TYPE_VALUE_LONG,
                           0 };

flowctrl func_file_upload(type_block* aktblock,opt* options,
                          joined_list<type_value>* args)
{
  PostdataBuilder postdata_builder;
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;
  back.returnval.stringval=0;

  char buf[MAX_LINE_SIZE];
  char buf1[MAX_LINE_SIZE];
  string separator,token;
  long maxsize=(*args)[0].val.longval;
  if (maxsize<0)
  {
    setvarulong("error",1);
    setvarstring("perror","file_upload: maxsize < 0");
    return back;
  };

  if (!fgets(buf,MAX_LINE_SIZE,stdin))
    {
      setvarulong("error",1);
      setvarstring("perror","file_upload: unexpected EOF");
      return back;
    };
  str_lfret(buf);
  separator=buf;
  string tmpfile;


  hashdir<char,string,256,0> declared;
  type_value* varptr;

  while ((const char*) buf!=separator+"--\r\n")
    {
      if (!fgets(buf,MAX_LINE_SIZE,stdin))
        {
          setvarulong("error",1);
          setvarstring("perror","file_upload: unexpected EOF");
          return back;
        };
      char* ptr=buf;
      string content_disposition=strsep(&ptr,";\n");
      string name=strsep(&ptr,";\n");
      string filename=strsep(&ptr,"\"");
      if (!defined(content_disposition) || !defined(name))
        {
          setvarulong("error",1);
          setvarstring("perror","file_upload: syntax error");
          return back;
        };

      /* go to data */
      buf[0]=0;
      while (strcmp(buf,"\r\n"))
	{
	  if (!fgets(buf,MAX_LINE_SIZE,stdin))
	    {
	      setvarulong("error",1);
	      setvarstring("perror","file_upload: unexpected EOF");
	      return back;
	    };
	};

      if (content_disposition!="Content-Disposition: form-data")
        {
          setvarulong("error",1);
          setvarstring("perror","file_upload: unknown disposition:"+
                       content_disposition);
          return back;
        };
      name=(string) (strchr(name.c_str(),'"')+1);
      strchr(name.lock(),'"')[0]=0;
      name.unlock();

      if (defined(filename) && strstr(filename.c_str()," filename=")==filename.c_str())
        {
          /* a file is transmitted */

	  /* set variable for filename */
	  filename=filename.c_str()+11;
	  //strstr(filename.zgr,"\"")[0]=0;
	  //filename=filename.zgr;
	  string varname="file";
	  type_value var;
	  var.type=TYPE_VALUE_STRING;
	  var.stringval=filename;
	  flowctrl call;
      call=declarevar(aktblock,options,&var,varname);
	  if (call.ctrl!=FLOW_OK)
	    {
	      setvarulong("error",1);
	      setvarstring("perror",
			   "file_upload: could not declare variable "+
			   varname+":"+call.error);
	      return back;
	    };

          unsigned long size=0;
          tmpfile=tmpnam(NULL);
          if (!defined(tmpfile))
            {
              setvarulong("error",1);
              setvarstring("perror","file_upload: tmpnam() failed");
              return back;
            };
          
          int handle;
          handle=open(tmpfile.c_str(),O_WRONLY|O_CREAT,
                      S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
          if (handle==-1)
            {
              setvarulong("error",1);
              setvarstring("perror","file_upload: cannot open "+tmpfile);
              return back;
            };
          long i;
          
          for (i=0;i!=MAX_LINE_SIZE;i++) { buf[i]=0; };
          if (!fgets(buf,MAX_LINE_SIZE,stdin))
            {
              setvarulong("error",1);
              setvarstring("perror","file_upload: unexpected EOF");
              return back;
            };
          while ((const char*) buf!=separator+"\r\n" &&
		 (const char*) buf!=separator+"--\r\n")
            {
              memcpy(buf1,buf,(size_t) MAX_LINE_SIZE);
              for (i=0;i!=MAX_LINE_SIZE;i++) { buf[i]=0; };
              if (!fgets(buf,MAX_LINE_SIZE,stdin))
                {
                  setvarulong("error",1);
                  setvarstring("perror","file_upload: unexpected EOF");
                  return back;
                };
              long aktsize=MAX_LINE_SIZE-1;
              for (i=MAX_LINE_SIZE-1;i>=0;i--)
                {
                  if (buf1[i])
		    {
		      if (buf1[i]=='\n')
			{
			  aktsize=i+1; break;
			}
		      else
			{
			  aktsize=MAX_LINE_SIZE-1; break;
			};
		    };
                };

              if ((const char*) buf==separator+"\r\n" ||
		  (const char*) buf==separator+"--\r\n")
                {
                  size+=(aktsize-2);
                  if (size>(unsigned long) maxsize)
                    {
                      setvarulong("error",1);
                      setvarstring("perror","file_upload: file too big");
                      return back;
                    };
                  if (write(handle,buf1,aktsize-2)!=aktsize-2)
                    {
                      setvarulong("error",1);
                      setvarstring("perror","file_upload: write() failed");
                      return back;
                    };
                  if (close(handle)==-1)
                    {
                      setvarulong("error",1);
                      setvarstring("perror","file_upload: close() failed");
                      return back;
                    };
                }
              else
                {
                  size+=aktsize;
                  if (size>(unsigned long) maxsize)
                    {
                      setvarulong("error",1);
                      setvarstring("perror","file_upload: file too big");
                      return back;
                    };
                  if (write(handle,buf1,aktsize)!=aktsize)
                    {
                      setvarulong("error",1);
                      setvarstring("perror","file_upload: write() failed");
                      return back;
                    };
                };
            };
        }
      else
        {
          /* normal form-data */
          string value="";
          if (!fgets(buf,MAX_LINE_SIZE,stdin))
            {
              setvarulong("error",1);
              setvarstring("perror","file_upload: unexpected EOF");
              return back;
            };
          while ((const char*) buf!=separator+"\r\n" &&
		 (const char*) buf!=separator+"--\r\n")
            {
              memcpy(buf1,buf,(size_t) MAX_LINE_SIZE);
              if (!fgets(buf,MAX_LINE_SIZE,stdin))
                {
                  setvarulong("error",1);
                  setvarstring("perror","file_upload: unexpected EOF");
                  return back;
                };
              if ((const char*) buf==separator+"\r\n" ||
		  (const char*) buf==separator+"--\r\n")
                {
                  str_lfret(buf1);
                  value=value+buf1;
                  
                  type_value var;
                  var.type=TYPE_VALUE_STRING;
                  var.stringval=value;
                  flowctrl call;
                  postdata_builder.append(name.c_str(), value.c_str());
                  call=declarevar(aktblock,options,&var,name);
                  if (call.ctrl!=FLOW_OK)
                    {
		      if (declared.get_listelem(name))
			    {
			       call=getvar(aktblock,options,varptr,name);
	                  if (call.ctrl!=FLOW_OK)
                            {
                              setvarulong("error",1);
                              setvarstring("perror",
                                           "postdata: hum ? could not get already defined variable "+name+":"+call.error);
                              back.returnval.val.ulongval=0;
                            }
                          else
                            {
                              varptr->stringval+=","+var.stringval;
                            };
                        }
                      else
                        {
                          setvarulong("error",1);
                          setvarstring("perror",
                                       "file_upload: could not declare variable "+
                                       name+":"+call.error);
                          return back;
                        };
                    }
                  else
                    {
                      char ch;
                      declared.insert_listelem(ch,name);
                    };
                }
              else
                {
                  value=value+buf1;
                };
            };
        };
    };

  type_value var;
  var.type=TYPE_VALUE_STRING;
  var.stringval=postdata_builder.get_result();
  flowctrl call;
  string name = "upload_postdata";
  call=declarevar(aktblock,options,&var, name);
    if (call.ctrl!=FLOW_OK)
      {
      if (declared.get_listelem(name))
	    {
	       call=getvar(aktblock,options,varptr,name);
              if (call.ctrl!=FLOW_OK)
                    {
                      setvarulong("error",1);
                      setvarstring("perror",
                                   "postdata: hum ? could not get already defined variable "+name+":"+call.error);
                      back.returnval.val.ulongval=0;
                    }
                  else
                    {
                      varptr->stringval+=","+var.stringval;
                    };
                }
              else
                {
                  setvarulong("error",1);
                  setvarstring("perror",
                               "file_upload: could not declare variable "+
                               name+":"+call.error);
                  return back;
                };
            }
          else
            {
              char ch;
              declared.insert_listelem(ch,name);
            };


  back.returnval.stringval=tmpfile;
  return back;
}

int args_htmlenc[]= { TYPE_VALUE_STRING,
                      0 };

flowctrl func_htmlenc(type_block* aktblock,opt* options,
                      joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;
  char buf[MAX_LINE_SIZE];

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","htmlenc: string not defined");
      return back;
    };

  const char* src=(*args)[0].stringval.c_str();
  back.returnval.stringval="";

  for (unsigned long i=0;src[i];i++)
    {
      if (src[i]=='<' || src[i]=='>' || src[i]=='\"' || src[i]=='\\' ||
          src[i]=='&')
        {
          snprintf(buf,MAX_LINE_SIZE,"&#%u;",(unsigned int) src[i]);
        }
      else
        {
          buf[0]=src[i];
          buf[1]=0;
        };

      back.returnval.stringval+=buf;
    };

  return back;
}

int args_phtmlenc[]= { TYPE_VALUE_STRING,
                       0 };

flowctrl func_phtmlenc(type_block* aktblock,opt* options,
                       joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=0;
  char buf[MAX_LINE_SIZE];

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","phtmlenc: string not defined");
      return back;
    };

  const char* src=(*args)[0].stringval.c_str();
  back.returnval.stringval="";

  for (unsigned long i=0;src[i];i++)
    {
      if (src[i]=='<' || src[i]=='>' || src[i]=='\"' || src[i]=='\\' ||
          src[i]=='&')
        {
          snprintf(buf,MAX_LINE_SIZE,"&#%u;",(unsigned int) src[i]);
        }
      else
        {
          buf[0]=src[i];
          buf[1]=0;
        };

      if (fwrite((void*) buf,1,strlen(buf),outstream)!=strlen(buf))
        {
          setvarulong("error",1);
          setvarstring("perror","phtmlenc: fwrite() failed");
          return back;
        };
    };

  back.returnval.val.ulongval=1;
  return back;
}

int args_fullhtmlenc[]= { TYPE_VALUE_STRING,
                          0 };

flowctrl func_fullhtmlenc(type_block* aktblock,opt* options,
                          joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","htmlenc: string not defined");
      return back;
    };

  back.returnval.stringval=urlencode((*args)[0].stringval.lock());

  return back;
}

