/*
  (c) 1998 netEstate (http://www.netestate.de/, info@netestate.de),
Michael Brunnbauer & Markus Hendel, Gesellschaft fuer Netzwerkkommunikation bR
Am Bluetenanger 8, 80995 Muenchen, Germany

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "wwwlib.h"

void urlprint(FILE* fh,url u)
{
  if (defined(u.scheme))
    { 
      fputs("Scheme: ",fh);
      fputs(u.scheme.c_str(),fh);
      fputs("\n",fh);
    };
  if (defined(u.scheme_specific))
    {
      fputs("Scheme-specific: ",fh);
      fputs(u.scheme_specific.c_str(),fh);
      fputs("\n",fh);
    };
  if (defined(u.login))
    {
      fputs("Login: ",fh);
      fputs(u.login.c_str(),fh);
      fputs("\n",fh);
    };
  if (defined(u.connection))
    {
      fputs("Connection: ",fh);
      fputs(u.connection.c_str(),fh);
      fputs("\n",fh);
    };
  if (defined(u.path))
    {
      fputs("Path: ",fh);
      fputs(u.path.c_str(),fh);
      fputs("\n",fh);
    };
  if (defined(u.user))
    {
      fputs("User: ",fh);
      fputs(u.user.c_str(),fh);
      fputs("\n",fh);
    };
  if (defined(u.pw))
    {
      fputs("Pw: ",fh);
      fputs(u.pw.c_str(),fh);
      fputs("\n",fh);
    };
  if (defined(u.host))
    {
      fputs("Host: ",fh);
      fputs(u.host.c_str(),fh);
      fputs("\n",fh);
    };
  if (defined(u.port))
    {
      fputs("Port: ",fh);
      fputs(u.port.c_str(),fh);
      fputs("\n",fh);
    };
  return;
};

url urlparse(char* text)
{
  url rueck;

  if (!strstr(text,":"))
    {
      /* not a valid url */
      return rueck;
    };

  /* get scheme */
  rueck.scheme=text;
  strstr(rueck.scheme.lock(),":")[0]=0;
  rueck.scheme.unlock();

  /* get scheme-specific part */
  rueck.scheme_specific=strstr(text,":")+1;
  text=rueck.scheme_specific.lock();

  /* strip double / */
  if (strstr(text,"//")==text)
    {
      text+=2;
    }
  else
    {
      return rueck;
    };

  /* strip optional login info */
  if (strstr(text,"@"))
    {
      rueck.login=text;
      strstr(rueck.login.lock(),"@")[0]=0;
      rueck.login.unlock();
      text=strstr(text,"@")+1;
    };

  /* divide into connection and path */
  rueck.connection=text;
  if (strstr(text,"/"))
    {
      strstr(rueck.connection.lock(),"/")[0]=0;
      rueck.connection.unlock();
      rueck.path=strstr(text,"/")+1;
    };

  /* divide connection */
  rueck.host=rueck.connection;
  if (strstr(rueck.connection.c_str(),":"))
    {
      strstr(rueck.host.lock(),":")[0]=0;
      rueck.host.unlock();
      rueck.port=strstr(rueck.connection.c_str(),":")+1;
    };

  /* divide login */
  if (defined(rueck.login))
    {
      rueck.user=rueck.login;
      if (strstr(rueck.login.c_str(),":"))
	{
	  strstr(rueck.user.lock(),":")[0]=0;
	  rueck.user.unlock();
	  rueck.pw=strstr(rueck.login.c_str(),":")+1;
	};
    };

  return rueck;
};

string urlencode(char* s)
{
  char buf[MAX_LINE_SIZE];
  unsigned char* src=(unsigned char*) s;

  string back="";
  
  if (!src) return 0;

  for (unsigned long i=0;src[i];i++)
    {
      if ((src[i]>=48 && src[i]<=57) ||
          (src[i]>=65 && src[i]<=90) ||
          (src[i]>=97 && src[i]<=122) ||
          src[i]==46)
        {
          buf[0]=src[i];
          buf[1]=0;
        }
      else
        {
          snprintf(buf,MAX_LINE_SIZE,"%%%02X",(unsigned int) src[i]);
        };
      back+=buf;
    };
  return back;
}

/* returncodes: 0 = EOF, 1 = OK, -1 = fehler */

int urlencode_parse(char** src,string& name,string& value)
{
  char buf[2];
  char hex[3];
  buf[1]=0;
  hex[2]=0;
  name="";value="";
  int c;

  /* read name */
  while ((**src)!='=')
    {
      switch (**src)
	{
	case 0:
	  {
	    log(LOG_ERROR,__FILE__,__LINE__,ERROR_SYNTAX_ERROR,
		"urlencode_parse: EOF while parsing name");
	    return -1 ;
	  };
	case '+':
	  {
	    name+=" ";
	    break;
	  };
	case '%':
	  {
            hex[0]=*(++(*src));
            if (!hex[0])
              {
                log(LOG_ERROR,__FILE__,__LINE__,ERROR_SYNTAX_ERROR,
                    "urlencode_parse: EOF after %");
                return -1;
              };
            hex[1]=*(++(*src));
            if (!hex[1])
	      {
		log(LOG_ERROR,__FILE__,__LINE__,ERROR_SYNTAX_ERROR,
		    "urlencode_parse: EOF after %");
		return -1;
	      };
	    if (sscanf(hex,"%x",&c)!=1)
	      {
		log(LOG_ERROR,__FILE__,__LINE__,ERROR_SYNTAX_ERROR,
		    "urlencode_parse: error with hex-character in name: %s",
		    (char*) hex);
		return -1;
	      };
	    buf[0]=(char) c;
	    name+=buf;
	    break;
	  };
	default:
	  {
	    buf[0]=**src;
	    name+=buf;
	    break;
	  };
	};
      (*src)++;
    };
  
  (*src)++;

  /* read value */
  while (**src && **src!='&')
    {
      switch (**src)
	{
        case '+':
          {
	    value+=" ";
	    break;
	  };
        case '%':
          {
            hex[0]=*(++(*src));
            if (!hex[0])
              {
                log(LOG_ERROR,__FILE__,__LINE__,ERROR_SYNTAX_ERROR,
                    "urlencode_parse: EOF after %");
                return -1;
              };
            hex[1]=*(++(*src));
            if (!hex[1])
              {
                log(LOG_ERROR,__FILE__,__LINE__,ERROR_SYNTAX_ERROR,
                    "urlencode_parse: EOF after %");
                return -1;
              };
            if (sscanf(hex,"%x",&c)!=1)
              {
                log(LOG_ERROR,__FILE__,__LINE__,ERROR_SYNTAX_ERROR,
                    "urlencode_parse: error with hex-character in value: %s",
                    (char*) hex);
                return -1;
              };
            buf[0]=(char) c;
            value+=buf;
            break;
          };
        default:
          {
	    buf[0]=**src;
	    value+=buf;
	    break;
          };
	};
      (*src)++;
    };

  if (**src=='&')
    {
      (*src)++;
      return 1;
    }
  else
    {
      return 0;
    };

  /* not reached */
  return 0;
}
